/*
 * Intercept open/stat calls and substitute filename with original filename
 * plus "-" and environment variable REV (if REV defined and such file
 * accessible).
 * 
 * To use:
 *  export LD_PRELOAD=libREV.so LD_LIBRARY_PATH=`pwd`:$LD_LIBRARY_PATH
 */

#define _GNU_SOURCE
#define __USE_GNU
#include <sys/syscall.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dlfcn.h>
#include <errno.h>
#include <stdarg.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>

int errno;

#define path_rev_size 256
char path_rev[path_rev_size];

void set_path_rev(const char* path) {
    char *rev;
    if ((rev = getenv("REV")) != NULL)
      if (strlen(rev) > 0)
	if (strlen(path) + strlen(rev) + 1 + 1 <= path_rev_size) {
	    sprintf(path_rev, "%s-%s\0", path, rev);
	    if (access(path_rev, F_OK) == 0)
		return;
	}
    strncpy(path_rev, path, path_rev_size);
    return;
}


int open(const char *pathname, int flags, ...) {
    static int (*realopen)(const char *, int flags, ...);
    if (!realopen) { realopen = dlsym(RTLD_NEXT, "open"); }
    set_path_rev(pathname);
/*    printf("open called: %s -> %s\n", pathname, path_rev);*/
    if (flags & O_CREAT) {
        va_list arg_list;
        mode_t mode;
        va_start(arg_list, flags);
        mode = va_arg(arg_list, mode_t);
        va_end(arg_list);
        return realopen(path_rev, flags, mode);
    } else {
	return realopen(path_rev, flags);
    }
}

int open64(const char *pathname, int flags, ...) {
    static int (*realopen64)(const char *, int flags, ...);
    if (!realopen64) { realopen64 = dlsym(RTLD_NEXT, "open64"); }
    set_path_rev(pathname);
/*    printf("open64 called: %s -> %s\n", pathname, path_rev);*/
    if (flags & O_CREAT) {
        va_list arg_list;
        mode_t mode;
        va_start(arg_list, flags);
        mode = va_arg(arg_list, mode_t);
        va_end(arg_list);
        return realopen64(path_rev, flags, mode);
    } else {
	return realopen64(path_rev, flags);
    }
}

int __open(const char *pathname, int flags, ...) {
    static int (*real__open)(const char *, int flags, ...);
    if (!real__open) { real__open = dlsym(RTLD_NEXT, "__open"); }
    set_path_rev(pathname);
/*    printf("__open called: %s -> %s\n", pathname, path_rev);*/
    if (flags & O_CREAT) {
        va_list arg_list;
        mode_t mode;
        va_start(arg_list, flags);
        mode = va_arg(arg_list, mode_t);
        va_end(arg_list);
        return real__open(path_rev, flags, mode);
    } else {
	return real__open(path_rev, flags);
    }
}

int __open64(const char *pathname, int flags, ...) {
    static int (*real__open64)(const char *, int flags, ...);
    if (!real__open64) { real__open64 = dlsym(RTLD_NEXT, "__open64"); }
    set_path_rev(pathname);
/*    printf("__open64 called: %s -> %s\n", pathname, path_rev);*/
    if (flags & O_CREAT) {
        va_list arg_list;
        mode_t mode;
        va_start(arg_list, flags);
        mode = va_arg(arg_list, mode_t);
        va_end(arg_list);
        return real__open64(path_rev, flags, mode);
    } else {
	return real__open64(path_rev, flags);
    }
}

FILE *fopen(const char *pathname, const char *mode) {
    static FILE* (*realfopen)(const char *, const char *);
    if (!realfopen) { realfopen = dlsym(RTLD_NEXT, "fopen"); }
    set_path_rev(pathname);
/*    printf("fopen called: %s -> %s\n", pathname, path_rev);*/
     return realfopen(path_rev, mode);
}

FILE *fopen64(const char *pathname, const char *mode) {
    static FILE* (*realfopen64)(const char *, const char *);
    if (!realfopen64) { realfopen64 = dlsym(RTLD_NEXT, "fopen64"); }
    set_path_rev(pathname);
/*    printf("fopen64 called: %s -> %s\n", pathname, path_rev);*/
     return realfopen64(path_rev, mode);
}

int stat(const char *file_name, struct stat *buf) {
    static int (*realstat)(const char *, struct stat *);
    if (!realstat) { realstat = dlsym(RTLD_NEXT, "stat"); }
    set_path_rev(file_name);
/*    printf("stat called: %s -> %s\n", file_name, path_rev);*/
    return realstat(path_rev, buf);
}

int stat64(const char *file_name, struct stat64 *buf) {
    static int (*realstat64)(const char *, struct stat64 *);
    if (!realstat64) { realstat64 = dlsym(RTLD_NEXT, "stat64"); }
    set_path_rev(file_name);
/*    printf("stat64 called: %s -> %s\n", file_name, path_rev);*/
    return realstat64(path_rev, buf);
}

int lstat(const char *file_name, struct stat *buf) {
    static int (*reallstat)(const char *, struct stat *);
    if (!reallstat) { reallstat = dlsym(RTLD_NEXT, "lstat"); }
    set_path_rev(file_name);
/*    printf("lstat called: %s -> %s\n", file_name, path_rev);*/
    return reallstat(path_rev, buf);
}

int lstat64(const char *file_name, struct stat64 *buf) {
    static int (*reallstat64)(const char *, struct stat64 *);
    if (!reallstat64) { reallstat64 = dlsym(RTLD_NEXT, "lstat64"); }
    set_path_rev(file_name);
/*    printf("lstat64 called: %s -> %s\n", file_name, path_rev);*/
    return reallstat64(path_rev, buf);
}

int __xstat(int ver, const char *file_name, struct stat *buf) {
    static int (*real__xstat)(int, const char *, struct stat *);
    if (!real__xstat) { real__xstat = dlsym(RTLD_NEXT, "__xstat"); }
    set_path_rev(file_name);
/*    printf("__xstat called: %s -> %s\n", file_name, path_rev);*/
    return real__xstat(ver, path_rev, buf);
}

int __xstat64(int ver, const char *file_name, struct stat64 *buf) {
    static int (*real__xstat64)(int, const char *, struct stat64 *);
    if (!real__xstat64) { real__xstat64 = dlsym(RTLD_NEXT, "__xstat64"); }
    set_path_rev(file_name);
/*    printf("__xstat64 called: %s -> %s\n", file_name, path_rev);*/
    return real__xstat64(ver, path_rev, buf);
}

int __lxstat(int ver, const char *file_name, struct stat *buf) {
    static int (*real__lxstat)(int, const char *, struct stat *);
    if (!real__lxstat) { real__lxstat = dlsym(RTLD_NEXT, "__lxstat"); }
    set_path_rev(file_name);
/*    printf("__lxstat called: %s -> %s\n", file_name, path_rev);*/
    return real__lxstat(ver, path_rev, buf);
}

int __lxstat64(int ver, const char *file_name, struct stat64 *buf) {
    static int (*real__lxstat64)(int, const char *, struct stat64 *);
    if (!real__lxstat64) { real__lxstat64 = dlsym(RTLD_NEXT, "__lxstat64"); }
    set_path_rev(file_name);
/*    printf("__lxstat64 called: %s -> %s\n", file_name, path_rev);*/
    return real__lxstat64(ver, path_rev, buf);
}

